(* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gontmodule.ml,v 1.10 2002/08/12 12:42:12 malekith Exp $
 *
 * Module implementation: Gontmodule
 * Description: We implement module interface here. As Gont modules are
 * very simple, so *this* module is :)
 *)

open Gonttree
open Gontutil

(* We serialize entire symbol. This is unsafe, as there can be cycle
 * typedef <-> symbol (and there is, in case of record types). However,
 * as it works (it is not guaranteed to work, by OCaml manual), we leave
 * it for now. *)
type mv_record = {mv_sym: symbol}

type mt_record = {mt_name: id;
                  mt_tydef: typedef}
                  
type mod_record = {mod_name: id;
                   mod_vals: mv_record list;
                   mod_types: mt_record list}

type module_info = {mi_name: string;
                    mi_used_exns: string list;
                    mi_used_modules: string list;
                    mi_init_fnc: string;
                    mi_fini_fnc: string}

(* We use it as magic. It is stored in all .gio files, if file contain
 * bad ident string -- it is rejected. This is probably not enough 
 * (we should also put version of gonttree.mli here, but... nothing's
 * perfect :^) *)
let ver = "Gont Binary, $Revision: 1.10 $"

let strip_module mod_name sym_name =
  let len = (String.length mod_name) + 2 in
  let slen = String.length sym_name in
  if slen > len && 
     (String.sub sym_name 0 len) = (mod_name ^ "::") then
    String.sub sym_name len (slen - len)
  else
    sym_name 

let save_module out_ch mod_name symbols types =
  let make_mv sym =
    assert (sym.sym_fnc_level == 0);
    assert (sym.sym_scope_level == 0);
    sym.sym_fnc_level <- -1;
    sym.sym_scope_level <- -1;
    {mv_sym = sym} in
  let make_mt ({td_name = n} as tydef) =
    {mt_name = strip_module mod_name n; 
     mt_tydef = {tydef with td_module = mod_name}} in
  let r = {mod_name = mod_name; 
           mod_vals = List.map make_mv symbols;
           mod_types = List.map make_mt types} in
  output_string out_ch ver;
  Marshal.to_channel out_ch r []

let save_module_info out_ch mi =
  output_string out_ch ver;
  Marshal.to_channel out_ch mi []
  
let save_module_info_list out_ch ldopts mil =
  output_string out_ch ver;
  Marshal.to_channel out_ch (ldopts, mil) []
;;

let modules : (string, mod_record) Hashtbl.t = Hashtbl.create 16

let iface_search_path = ref []

let reset () = Hashtbl.clear modules

exception IO_error of string

let set_iface_search_path p =
  iface_search_path := p
  
open Unix

(* Header:
 * "!<gont-arch>\nSSSSSSSSSSSSSSSSSSSS"
 *  1234567890123 45678901234567890123
 * where S...S (there is 20 of them) is the size of archive data, excluding 
 * header and trailing index. *)

type file_header = {
  fh_pos: int;
  fh_size: int;
}

type arch_index = {
  ai_files: (string * file_header) list;
}

let magic = "!<gont-arch>\n"

let blocking_read fd len =
  let rec read_rec fd ret off len =
    let rd = read fd ret off len in
    if rd <= 0 then
      raise (IO_error "file too short")
    else
      if rd < len then
        read_rec fd ret (off + rd) (len - rd)
      else
        ret in
  let str = String.make len ' ' in
  read_rec fd str 0 len 

let blocking_write fd str =
  let rec write_rec fd str off len =
    let wr = write fd str off len in
    if wr <= 0 then
      raise (IO_error "short write")
    else
      if wr < len then
        write_rec fd str (off + wr) (len - wr)
      else
        () in
   write_rec fd str 0 (String.length str)

let raise_io_error (errno, arg1, arg2) =
  raise (IO_error (arg1 ^ ": " ^ (error_message errno) ^ 
                     " (" ^ arg2 ^ ")"))
  
let make_archive target files =
  try
    let write_number fd k =
      blocking_write fd (xf "%020d" k) in
    let fd = openfile target [O_WRONLY; O_TRUNC; O_CREAT] 0o666 in
    (* We don't know the size yet. *)
    let _ = blocking_write fd magic in
    let _ = write_number fd 0 in
    let rec write_files idx off files =
      match files with
      | x::xs -> 
        let rfd = openfile x [O_RDONLY] 0 in
        let st = fstat rfd in
        if st.st_size == 0 then
          raise (IO_error ("cowardly refusing to add empty file (" ^ x ^ 
                           ") to archive"))
        else
          let content = blocking_read rfd st.st_size in
          let _ = blocking_write fd content in
          write_files ((x, {fh_pos = off; fh_size = st.st_size}) :: idx) 
                      (off + st.st_size)
                      xs
      | [] -> (off, idx) in
    let (sz, idx) = write_files [] 0 files in
    let idx = Marshal.to_string {ai_files = idx} [] in
    write_number fd (String.length ver);
    blocking_write fd ver;
    write_number fd (String.length idx);
    blocking_write fd idx;
    let _ = lseek fd (String.length magic) SEEK_SET in
    write_number fd sz;
    close fd
  with Unix_error (x, y, z) -> raise_io_error (x, y, z)
;;

let try_extract lib member =
  let read_number fd =
    let str = blocking_read fd 20 in
    int_of_string str in
  let fd = openfile lib [O_RDONLY] 0 in
  try
    let read_magic = blocking_read fd 13 in
    if read_magic <> magic then
      raise (IO_error "bad magic")
    else
      let sz = read_number fd in
      let _ = lseek fd sz SEEK_CUR in
      let idx_magic_sz = read_number fd in
      let idx_magic = blocking_read fd idx_magic_sz in
      let idx_sz = read_number fd in
      let idx = blocking_read fd idx_sz in
      if idx_magic <> ver then
        raise (IO_error "bad version of archive, please remake")
      else
        let idx = (Marshal.from_string idx 0 : arch_index) in
        try
          let fh = List.assoc member idx.ai_files in
          let _ = lseek fd (fh.fh_pos + 20 + String.length magic) SEEK_SET in
          let ret = blocking_read fd fh.fh_size in
          let _ = close fd in
          ret
        with Not_found -> raise (Unix_error (ENOENT, "try_extract", member))
  with e -> 
    let _ = close fd in
    raise e
  
let content_of_filename name =
  try
    let st = Unix.stat name in
    if st.st_size == 0 then
      raise (IO_error "file has zero size")
    else
      let fd = openfile name [O_RDONLY] 0 in
      let ret = blocking_read fd st.st_size in
      let _ = close fd in
      ret
  with Unix_error (x, y, z) -> raise_io_error (x, y, z)

let rec read_file path name =
  match path with
  | [] -> raise (IO_error "file not found")
  | x::xs ->
    try
      let st = stat x in
      match st.st_kind with
      | S_DIR ->
        let name = Filename.concat x name in
        let _ = Unix.stat name in       (* To raise ENOENT *)
        content_of_filename name
      | _ ->
        try_extract x name
    with 
    | Unix_error (ENOENT, _, _) ->
        (* Skip absent files. *)
        read_file xs name
    | Unix_error (x, y, z) -> raise_io_error (x, y, z)

(* Convert name of module to filesystem name. *)
let iface_name m = (String.uncapitalize m) ^ ".gio"

let module_info_of_string content m =
  let ver_len = String.length ver in
  if String.sub content 0 ver_len <> ver then
    raise (Error (xf "module implementation file for %s was compiled \
                      with different version, please recompile" m))
  else
    try
      (Marshal.from_string content ver_len : module_info)
    with End_of_file ->
      raise (Error (xf "unexpected end of file, reading module \
                        implementation for `%s'" m))

let module_info_list_of_string content f =
  let ver_len = String.length ver in
  if String.sub content 0 ver_len <> ver then
    raise (Error (xf "module implementation library `%s' was compiled \
                      with different version, please recompile" f))
  else
    try
      (Marshal.from_string content ver_len : (string * module_info list))
    with End_of_file ->
      raise (Error (xf "unexpected end of file, reading module \
                        library `%s'" f))

let module_info_of_file f =
  module_info_of_string (content_of_filename f) f

let module_info_list_of_file f =
  module_info_list_of_string (content_of_filename f) f

let open_module m =
  let load m =
    let name = iface_name m in
    try
      let content = read_file !iface_search_path name in
      let ver_len = String.length ver in
      if String.sub content 0 ver_len <> ver then
        raise (Error (xf "module interface file for %s was compiled \
                          with different version, please recompile \
                          the interface" m))
      else
        try
          let mr = (Marshal.from_string content ver_len : mod_record) in
          if mr.mod_name <> m then
            raise (Error (xf "interface file contains module %s, \
                              instead of expected %s" mr.mod_name m))
          else
             Hashtbl.add modules mr.mod_name mr
        with End_of_file ->
          raise (Error (xf "unexpected end of file, reading module interface \
                            for `%s'" m))
    with IO_error s ->
      raise (Error (xf "cannot find module interface for %s: %s" m s)) in
  try
    ignore(Hashtbl.find modules m)
  with Not_found -> load m
;;

let find_module m =
  open_module m;
  Hashtbl.find modules m

exception Symbol_found of symbol
exception Tydef_found of typedef

let get_symbol m =
   match m with
   | {mv_sym = sym} ->
       sym
;;

let import_symbol kind mod_name sym_name =
  let qn = mod_name ^ "::" ^ sym_name in
  try
    List.iter (fun s -> 
       match s with
       | {mv_sym = {sym_name = n;
                    sym_kind = k}} when (n = qn && kind = k) ->
            raise (Symbol_found (get_symbol s))
       | _ -> ()) (find_module mod_name).mod_vals;
    raise Not_found
  with Symbol_found s -> s
  
let import_type_symbol mod_name sym_name =
  try
    List.iter (fun s -> 
       match s with
          {mt_name = n; mt_tydef = t} when (n = sym_name) -> 
            raise (Tydef_found t)
        | _ -> ()) (find_module mod_name).mod_types;
    raise Not_found
  with Tydef_found t -> t
    
let globals mod_name =
  let m = find_module mod_name in
  ((List.map get_symbol m.mod_vals),
   (List.map (fun {mt_tydef = t} -> t) m.mod_types))
;;
